#!/usr/bin/env python

import copy
import math
import numpy.random as dist
import StringIO

import unittest

import schedcat.generator.tasks as tg
import schedcat.generator.tasksets as tsgen

def const(v):
    def _draw():
        return v
    return _draw

class node:
    def __init__(self, id=-1, level=-1, graph = None):
        self.id = id
        self.graph = graph
        self.isSrc = False
        self.isSink = False
        self.isSpine = False
        # list of predecessor nodes
        self.pred = []
        # list of successor nodes
        self.succ = []
        # list of inbound edges
        self.inEdges = []
        # list of outbound edges
        self.outEdges = []

        # variables used during generation
        self.privLevel = level

    def isAncestor(self, other):
        q = copy.copy(self.pred)
        while q:
            v = q.pop()
            if v == other:
                return True
            q.extend(v.pred)
        return False

    def isDescendant(self, other):
        return other.isAncestor(self)

    def degreeIn(self):
        return len(self.inEdges)

    def degreeOut(self):
        return len(self.outEdges)

    def degree(self):
        return len(self.inEdges) + len(self.outEdges)

    def __repr__(self):
        graph_id = self.graph.id if self.graph and hasattr(self.graph, 'id') else -1
        stem = 'node_%s(gid:%d,l:%d,src:%d,sink:%d,spine:%d)' % (self.id, graph_id, self.privLevel, self.isSrc, self.isSink, self.isSpine)
        pred_str = 'preds{'
        for n in self.pred:
            pred_str += str(n.id) + ','
        pred_str += '}'
        succ_str = 'succs{'
        for n in self.succ:
            succ_str += str(n.id) + ','
        succ_str += '}'
        return '%s %s %s' % (stem, pred_str, succ_str)

    def dot(self):
        attr = []
        if self.isSrc or self.isSink:
            attr.append('shape=doublecircle')
        else:
            attr.append('shape=circle')
        if not (self.isSrc or self.isSink):
            attr.append('style=""')
            attr.append('color=""')
        elif self.isSrc:
            attr.append('style=filled')
            attr.append('color="#56A0D3"')
        elif self.isSink:
            attr.append('style=filled')
            attr.append('color="#E04050D3"')
        node_str = 'node [%s]; %d;' % (', '.join(attr), self.id)
        if hasattr(self, 'task'):
            # assumes time is in microseconds
            cost_str = format(self.task.cost/1000.0, '.4f').rstrip('0').rstrip('.')
            period_str = format(self.task.period/1000.0, '.4f').rstrip('0').rstrip('.')
            extra_str = ' /* node_%d: exec: %s, period: %s, split: %d, cluster %d, is_src: %s, is_sink: %s */' % (
                self.id, cost_str, period_str, self.task.split, self.task.partition, self.isSrc, self.isSink)
        else:
            extra_str = ''
        return node_str + extra_str

class edge:
    def __init__(self, a, b):
        self.p = a
        self.s = b
        self.wss = 0

    def __repr__(self):
        return 'node_%d -> node_%d' % (self.p.id, self.s.id)

    def dot(self):
        criticalPath = (self.p.isSpine and self.s.isSpine and abs(self.s.privLevel - self.p.privLevel) == 1)
        edge_str = '%d -> %d [%s];' % (self.p.id, self.s.id, ('style="", color=dimgrey', 'style=bold, color="#E04050"')[criticalPath])
        wss_str = format(self.wss, '.4f').rstrip('0').rstrip('.')
        extra_str = ' /* parent: node_%d, child: node_%d, wss: %s, is_spine: %s */' % (self.p.id, self.s.id, wss_str, criticalPath)
        return edge_str + extra_str

class graph:
    def __init__(self):
        self.nodes = []
        self.sources = []
        self.sinks = []
        self.edges = []
        self.nodesAtLevel = {}
        self.depth = 0
        # assume a sporadic release from sources
        self.isSporadic = True

    def __repr__(self):
        s = ''
        for i in self.nodes:
            n = str(i)
            s += n + '\n'
        return s

    def dot(self):
        outs = StringIO.StringIO()
        outs.write('digraph G {\n')
        outs.write('\trankdir=LR;\n')
        outs.write('\tratio="compress";\n')
        outs.write('\tsize="11,8.5";\n')

        graph_name = 'graph' if not hasattr(self, 'id') else 'graph_%d' % self.id
        extra_str = '/* %s: depth: %d, nr_nodes: %d, nr_src: %d, nr_sink: %d, nr_edges: %d */' % (
            graph_name, self.depth, len(self.nodes), len(self.sources), len(self.sinks), len(self.edges))
        outs.write('\t%s\n' % extra_str)

        # draw nodes
        for n in self.nodes:
            outs.write('\t')
            outs.write(n.dot())
            outs.write('\n')

        # draw edges
        for e in self.edges:
            outs.write('\t')
            outs.write(e.dot())
            outs.write('\n')

        orphans = [n for n in self.nodes if not n.succ]
        for o in orphans:
            outs.write('\t%d;\n' % o.id)

        # end the graph
        outs.write('}')
        return outs.getvalue()

def bound_graph_latency(g):
    sporadic_latency = 0
    graph_latency = 0

    if len(g.nodes) == 1:
        sporadic_latency = g.nodes[0].task.response_time
    else:
        # We assume all nodes share a period
        period = g.nodes[0].task.period

        for n in g.nodes:
            n.latency = 0 if not n.isSrc else max(period, n.task.response_time)
            n.isQueued = False
        for e in g.edges:
            e.longest = False

        # accumulate latencies down the graph
        queue = g.sources[:]
        while len(queue) != 0:
            # breadth-first propagation of latencies from srcs to sinks
            v = queue.pop(0)
            v.isQueued = False
            for e in v.outEdges:
                latency = v.latency + max(period, e.s.task.response_time)
                # if we updated the latency, then we need to revisit the node
                if latency > e.s.latency:
                    # clear out old longest
                    cur_longest = [e for e in e.s.inEdges if e.longest == True]
                    if len(cur_longest) > 0:
                        assert len(cur_longest) == 1
                        cur_longest[0].longest = False
                    # set new longest
                    e.longest = True
                    e.s.latency = latency
                    if e.s.isQueued == False:
                        e.s.isQueued = True
                        queue.append(e.s)

        max_sink = max(g.sinks, key=lambda n: n.latency)
        longest_path = []
        longest_path.append(max_sink)
        queue.append(max_sink)
        while len(queue) != 0:
            v = queue.pop(0)
            if v.inEdges:
                longest_edge = [e for e in v.inEdges if e.longest is True]
                assert len(longest_edge) == 1
                longest_edge = longest_edge[0]
                # prepend to the path
                longest_path.insert(0, longest_edge.p)
                queue.append(longest_edge.p)
            else:
                assert v.isSrc
        assert len(longest_path) <= g.depth

        depth_latency = len(longest_path) * period
        sporadic_latency = max_sink.latency - depth_latency

        if g.isSporadic:
            # factor the accumlated latency into grapth-depth and
            # tardiness-based components

            if max_sink.task.response_time < period:
                # Optimization:
                #  * Remove one period from depth_latency
                #  * Add response time of sink to depth_latency
                # (We don't modify sporadic_latency since we know the sink's
                # contribition was 0.)
                graph_latency = depth_latency - period + max_sink.task.response_time
            else:
                graph_latency = depth_latency
        else:
            # TODO: Optimize of the sink tasks response time
            # Rate-based bound:
            graph_latency = 4 * depth_latency

    assert sporadic_latency >= 0
    assert graph_latency >= 0

#    print 'depth:           \t',g.depth
#    print 'period:          \t',period
#    print 'd*p:             \t',g.depth * period
#    print 'combined:        \t', sporadic_latency + graph_latency
#    print 'sporadic latency:\t', sporadic_latency
#    print 'graph latency:   \t', graph_latency
#    raw_input('press enter')

    return sporadic_latency + graph_latency

# tighter bound w/o 'max-period'
def bound_graph_latency2(g):
    if len(g.nodes) == 1:
        return g.nodes[0].task.response_time
    else:
        # We assume all nodes share a period
        period = g.nodes[0].task.period

        for n in g.nodes:
            n.latency = 0 if not n.isSrc else max(period, n.task.response_time)
            n.isQueued = False
        for e in g.edges:
            e.longest = False

        # accumulate latencies down the graph
        queue = g.sources[:]
        while len(queue) != 0:
            # breadth-first propagation of latencies from srcs to sinks
            v = queue.pop(0)
            v.isQueued = False
            for e in v.outEdges:
                # only sum up the task response time
                latency = v.latency + e.s.task.response_time
                # if we updated the latency, then we need to revisit the node
                if latency > e.s.latency:
                    # clear out old longest
                    cur_longest = [e for e in e.s.inEdges if e.longest == True]
                    if len(cur_longest) > 0:
                        assert len(cur_longest) == 1
                        cur_longest[0].longest = False
                    # set new longest
                    e.longest = True
                    e.s.latency = latency
                    if e.s.isQueued == False:
                        e.s.isQueued = True
                        queue.append(e.s)

        max_sink = max(g.sinks, key=lambda n: n.latency)
        return max_sink.latency

def compute_ideal_graph_latency(g):
    graph_latency = 0
    if len(g.nodes) == 1:
        graph_latency = g.nodes[0].task.cost
    else:
        for n in g.nodes:
            n.latency = 0 if not n.isSrc else n.task.cost
            n.isQueued = False

        queue = g.sources[:]

        while len(queue) != 0:
            # breadth-first propagation of latencies from srcs to sinks
            v = queue.pop(0)
            v.isQueued = False
            for e in v.outEdges:
                latency = v.latency + e.s.task.cost
                # if we updated the latency, then we need to revisit the node
                if latency > e.s.latency:
                    e.s.latency = latency
                    if e.s.isQueued == False:
                        e.s.isQueued = True
                        queue.append(e.s)
        graph_latency = max(g.sinks, key=lambda n: n.latency).latency
    assert graph_latency > 0
    return graph_latency

def compute_hrt_ideal_graph_latency(g):
    graph_latency = 0
    if len(g.nodes) == 1:
        graph_latency = g.nodes[0].task.deadline
    else:
        for n in g.nodes:
            n.latency = 0 if not n.isSrc else n.task.deadline
            n.isQueued = False

        queue = g.sources[:]

        while len(queue) != 0:
            # breadth-first propagation of latencies from srcs to sinks
            v = queue.pop(0)
            v.isQueued = False
            for e in v.outEdges:
                latency = v.latency + e.s.task.deadline
                # if we updated the latency, then we need to revisit the node
                if latency > e.s.latency:
                    e.s.latency = latency
                    if e.s.isQueued == False:
                        e.s.isQueued = True
                        queue.append(e.s)
        graph_latency = max(g.sinks, key=lambda n: n.latency).latency
    assert graph_latency > 0
    return graph_latency

def link(up, down):
    up.succ.append(down)
    down.pred.append(up)

def eligible_predecessor_of(n, maxDeg = None):
    def _check(candidate):
        if candidate.isSink:
            return False
        if maxDeg != None and maxDeg != 0 and len(candidate.succ) >= maxDeg:
            return False
        if candidate in n.pred:
            return False
        return True
    return _check

def pick_predecessor(g, fromLevel, leap_picker, is_eligible_predecessor):
    nStepsBack = min(max(1, leap_picker()), fromLevel)
    targetLevel = fromLevel - nStepsBack
    for l in range(targetLevel,-1,-1):
        candidates = [x for x in g.nodesAtLevel[l] if is_eligible_predecessor(x)]
        if not candidates:
            continue
        return tg.uniform_choice(candidates)()
    return None

def is_disjoint(g):
    for n in g.nodes:
        n.visited = False
    # visit all nodes reachable from main source
    q = [g.nodes[0]]
    while q:
        n = q.pop()
        n.visited = True
        for s in n.succ:
            if not s.visited:
                q.append(s)
    # visit all nodes reachable from main sink
    q = [g.nodes[1]]
    while q:
        n = q.pop()
        n.visited = True
        for p in n.pred:
            if not p.visited:
                q.append(p)
    # any unvisited nodes?
    if [n for n in g.nodes if not n.visited]:
        return True
    return False

def makeSpine(nrNodes, depth):
    g = graph()
    g.depth = depth
    for i in range(nrNodes):
        g.nodes.append(node(i, graph = g))
    for i in range(depth):
        g.nodesAtLevel[i] = []

    base = 0
    # insert main src
    g.nodesAtLevel[0].append(g.nodes[base])
    g.nodes[base].privLevel = 0
    g.nodes[base].isSrc = True
    g.nodes[base].isSpine = True
    g.sources.append(g.nodes[base])
    if nrNodes == 1:
        g.nodes[base].isSink = True
        g.sinks.append(g.nodes[base])
        return g
    base = base + 1

    # insert main sink
    g.nodesAtLevel[depth-1].append(g.nodes[base])
    g.nodes[base].privLevel = depth-1
    g.nodes[base].isSink = True
    g.nodes[base].isSpine = True
    g.sinks.append(g.nodes[base])
    base = base + 1

    # insert the spine
    for i in range(base, base + depth - 2):
        l = i - base + 1
        g.nodesAtLevel[l].append(g.nodes[i])
        g.nodes[i].privLevel = l
        g.nodes[i].isSpine = True

    # set up the critical path
    for i in range(0, depth-1):
        link(g.nodesAtLevel[i][0], g.nodesAtLevel[i+1][0])
    return g

def assignExtraSources(g, depth, nrSrc):
    candidates = []
    # pick remaining srcs at random from nodes < depth/2
    for l in range(0, depth/2):
        for n in g.nodesAtLevel[l]:
            if not (n.isSrc or n.isSink or n.isSpine):
                candidates.append(n)
    if len(candidates) < nrSrc:
        raise Exception('not enough nodes available for sources')
    for i in range(nrSrc):
        which = tg.uniform_int(0, len(candidates)-1)()
        candidates[which].isSrc = True
        g.sources.append(candidates[which])
        del candidates[which]
    return g

def assignExtraSinks(g, depth, nrSink):
    candidates = []
    # pick remaining srcs at random from nodes < depth/2
    for l in range(int(depth/2.0+1.0), depth):
        for n in g.nodesAtLevel[l]:
            if not (n.isSrc or n.isSink or n.isSpine):
                candidates.append(n)
    if len(candidates) < nrSink:
        raise Exception('not enough nodes available for sinks')
    for i in range(nrSink):
        which = tg.uniform_int(0, len(candidates)-1)()
        candidates[which].isSink = True
        g.sinks.append(candidates[which])
        del candidates[which]
    return g

def flipOrientation(g):
    depth = len(g.nodesAtLevel)
    for n in g.nodes:
        n.pred, n.succ = n.succ, n.pred
        n.isSrc, n.isSink = n.isSink, n.isSrc
    g.sources, g.sinks = g.sinks, g.sources
    temp = {}
    for l in g.nodesAtLevel:
        temp[depth-l] = g.nodesAtLevel[l]
    g.nodesAtLevel = temp
    return g

def finalizeEdges(g):
    # finally, add the actual edges
    for p in g.nodes:
        for s in p.succ:
            e = edge(p, s)
            g.edges.append(e)
            p.outEdges.append(e)
            s.inEdges.append(e)
    return g

class GraphGenerator(object):
    """PGM graph generator"""

    def __init__(self, nrNodes, depth, levelDist,
                 fanDist, oppFanCap, edgeDist,
                 nrSrc, nrSink, bottomUp, flip):
        self.nrNodes = nrNodes
        self.depth = depth
        self.levelDist = levelDist
        self.fanDist = fanDist
        self.oppFanCap = oppFanCap
        self.edgeDist = edgeDist
        self.nrSrc = nrSrc
        self.nrSink = nrSink
        self.bottomUp = bottomUp
        self.flip = flip

    def __placeNodes(self, g, nrNodes, depth):
        # distribute remaining nodes among the middle levels
        for i in range(depth, nrNodes):
            # we need to adjust our level distro the actual depth of the graph
            l = self.levelDist.getadjusted(1, max(1,depth-2))()
            if self.flip:
                # transpose the level across the center level of the graph
                l = depth - l - 1
            g.nodes[i].privLevel = l
            g.nodesAtLevel[l].append(g.nodes[i])
        return g

    def __lockdownPlacement(self, g, depth):
        # set initial edges that lock nodes into their selected level
        for n in g.nodes:
            if n.isSrc or n.pred:
                continue
            p = pick_predecessor(g, n.privLevel, const(1), eligible_predecessor_of(n, self.oppFanCap))
            if not p:
                p = pick_predecessor(g, n.privLevel, const(1), eligible_predecessor_of(n))
                assert p
            link(p, n)
        return g

    def __addExtraEdges(self, g, depth):
        # compute size of graph at each level, excluding sinks
        count = [0]*depth
        for l in range(depth):
            if l != 0:
                count[l] = count[l-1]
            count[l] += len(g.nodesAtLevel[l])
            for n in g.nodesAtLevel[l]:
                if n.isSink:
                    count[l] -= 1
        # generate edges, skipping the first level (which is reserved for the main source)
        for n in g.nodes:
            if n.isSrc:
                continue
            l = n.privLevel
            deep = max(0, l - self.edgeDist.min)
            shallow = max(0, l - self.edgeDist.max)
            nrPredCap = count[deep]
            if shallow != deep:
                nrPredCap -= count[shallow]
            degree = min(nrPredCap, self.fanDist())
            for i in range(len(n.pred), degree):
                p = pick_predecessor(g, n.privLevel, self.edgeDist, eligible_predecessor_of(n, self.oppFanCap))
                if p:
                    link(p, n)
                else:
                    break
        # make sure all non-sink nodes have a successor
        toFix = [n for n in g.nodes if not (n.succ or n.isSink)]
        for n in toFix:
            # pick a random sink node at a depth greater than our own.
            # the main sink node will always be a candidate.
            candidates = [m for m in g.sinks if m.privLevel > n.privLevel]
            if not candidates:
                raise Exception('Could not find sink.')
            which = tg.uniform_choice(candidates)()
            link(n, which)
        return g

    def __graph(self):
        """Make a graph"""
        nrNodes = self.nrNodes()

        assert nrNodes != 0

        depth = self.depth()

        # cannot be deeper than the total number of nodes
        if nrNodes < depth:
            depth = nrNodes
        if depth == 1 and nrNodes > 1:
            depth += 1
        # two levels are reserved for the main src and sink, thus
        # depth must be at least 3 for graphs with #nodes > 2
        if nrNodes > 2 and depth <= 2:
            depth += 1

        g = makeSpine(nrNodes, depth)
        if nrNodes == 1:
            return g

        g = self.__placeNodes(g, nrNodes, depth)

        # handle extra sources and sinks
        nrSrc = self.nrSrc()
        nrSink = self.nrSink()
        if (self.bottomUp):
            nrSrc, nrSink = nrSink, nrSrc
        if nrSrc > 1:
            g = assignExtraSources(g, depth, nrSrc-1)
        if nrSink > 1:
            g = assignExtraSinks(g, depth, nSink-1)

        g = self.__lockdownPlacement(g, depth)
        g = self.__addExtraEdges(g, depth)

        # there might be disjoint graphs in combined multi-src/multi-sink graphs
        # for now, just detect the case and raise an exception
        if nrSrc != 1 and nrSink != 1 and is_disjoint(g):
            raise Exception('Disjoint graph')

        # re-orient the graph
        if not self.bottomUp:
            g = flipOrientation(g)

        g = finalizeEdges(g)

        return g

    def graph(self, maxAttempts = 100):
        lastExcept = None
        for i in range(maxAttempts):
            try:
                return self.__graph()
            except Exception as e:
                lastExcept = e
        msg = 'Failed to generate graph after %d attempts.' % maxAttempts
        if lastExcept:
            msg += ' Last generator exception: ' + str(lastExcept)
        raise Exception(msg + '\n')

class ParallelPipeGraphGenerator(object):
    """Generates PGM graphs of parallel pipelines."""
    def __init__(self, nrNodes, depth):
        self.nrNodes = nrNodes
        self.depth = depth

    def __graph(self):
        nrNodes = self.nrNodes()
        depth = self.depth()

        if nrNodes < depth:
            depth = nrNodes
        if depth == 1 and nrNodes > 1:
            depth += 1
        if nrNodes > 2 and depth <= 2:
            depth += 1

        g = makeSpine(nrNodes, depth)
        if nrNodes == 1:
            return g

        nrToAdd = nrNodes - depth

        base = depth
        nrAdded = 0

        while nrAdded < nrToAdd:
            max_depth = min(nrToAdd - nrAdded, max(1, depth-2))
            leg_depth = tg.uniform_int(1, max_depth)()

            offset = base + nrAdded

            firstNode = g.nodes[offset]
            lastNode = None

            for i in range(leg_depth):
                node = g.nodes[offset+i]
                g.nodesAtLevel[i+1].append(node)
                node.privLevel = i+1
                if lastNode:
                    link(lastNode, node)
                lastNode = node
            nrAdded += leg_depth
            # link into the src and sink
            link(g.sources[0], firstNode)
            link(lastNode, g.sinks[0])

        if is_disjoint(g):
            raise Exception('Disjoint graph')

        g = finalizeEdges(g)

        return g

    def graph(self):
        return self.__graph()


class RectGraphGenerator(object):
    """Generates rectangular PGM graphs."""
    def __init__(self, nrNodes, aspectRatio):
        """Aspect ratio is defined by width/depth. (width == fan-out of source node)"""
        self.nrNodes = nrNodes
        self.aspectRatio = aspectRatio

    def __graph(self):
        nrNodes = self.nrNodes()
        ratio = self.aspectRatio()

        depth = max(1, int(math.sqrt(nrNodes/ratio)))

        if nrNodes < depth:
            depth = nrNodes
        if depth == 1 and nrNodes > 1:
            depth += 1
        if nrNodes > 2 and depth <= 2:
            depth += 1

        g = makeSpine(nrNodes, depth)
        if nrNodes == 1:
            return g

        width = int(nrNodes / depth)
        extra = nrNodes - depth*width

        legs = [depth]*(width-1)
        if extra != 0:
            legs.append(extra)

        nrToAdd = nrNodes - depth
        base = depth
        nrAdded = 0
        for l in legs:
            leg_depth = l
            offset = base + nrAdded
            firstNode = g.nodes[offset]
            lastNode = None
            for i in range(leg_depth):
                node = g.nodes[offset+i]
                g.nodesAtLevel[i].append(node)
                node.privLevel = i
                if lastNode:
                    link(lastNode, node)
                lastNode = node
            nrAdded += leg_depth
            # link into the src and sink
            link(g.sources[0], firstNode)
            link(lastNode, g.sinks[0])

        if is_disjoint(g):
            raise Exception('Disjoint graph')

        g = finalizeEdges(g)

        return g

    def graph(self):
        return self.__graph()

class RectGraphGenerator2(object):
    """Generates rectangular PGM graphs."""
    def __init__(self, nrNodes, width):
        """Aspect ratio is defined by width/depth. (width == fan-out of source node)"""
        self.nrNodes = nrNodes
        self.width = width

    def __graph(self):
        nrNodes = self.nrNodes()
        width = self.width()

        # maximum possible width
        if width > nrNodes-2:
            width = max(1, nrNodes-2)

        depth = int(max(1, nrNodes-2)/width)

        legs = [depth] * width
        remainder = nrNodes - 2 - sum(legs)
        while remainder > 0:
            assert len(legs) > 0
            for i,l in enumerate(legs):
                legs[i] += 1
                remainder -= 1
                if remainder == 0:
                    break

        spinelen = min(legs[0]+2, nrNodes)
        g = makeSpine(nrNodes, spinelen)
        if nrNodes == 1:
            return g
        legs.pop(0)

        nrToAdd = sum(legs)
        base = spinelen
        nrAdded = 0
        for l in legs:
            leg_depth = l
            offset = base + nrAdded
            firstNode = g.nodes[offset]
            lastNode = None
            for i in range(leg_depth):
                node = g.nodes[offset+i]
                lvl = i+1
                g.nodesAtLevel[lvl].append(node)
                node.privLevel = lvl
                if lastNode:
                    link(lastNode, node)
                lastNode = node
            nrAdded += leg_depth
            # link into the src and sink
            link(g.sources[0], firstNode)
            link(lastNode, g.sinks[0])

        if is_disjoint(g):
            raise Exception('Disjoint graph')

        g = finalizeEdges(g)

        return g

    def graph(self):
        return self.__graph()

class TreeGraphGenerator(object):
    """Generates tree-based PGM graphs. All interior nodes have one parent."""
    def __init__(self, nrNodes, nodeDegree):
        self.nrNodes = nrNodes
        self.nodeDegree = nodeDegree

    def __graph(self):
        nrNodes = self.nrNodes()
        degree = self.nodeDegree()

        tree_height = int(math.ceil(math.log(nrNodes, degree)))

        g = graph()
        for i in range(nrNodes):
            g.nodes.append(node(i, graph = g))
        for i in range(tree_height+1):
            g.nodesAtLevel[i] = []

        # make the src first
        g.nodesAtLevel[0].append(g.nodes[0])
        g.nodes[0].privLevel = 0
        g.nodes[0].isSrc = True
        g.sources.append(g.nodes[0])
        if nrNodes == 1:
            g.node[0].isSink = True
            g.sinks.append(g.nodes[0])
            return g

        for i in range(nrNodes-1):
            # add children for node i.
            parent = g.nodes[i]
            children = []
            for j in range(degree):
                # see forumla for embedding trees in arrays
                idx = degree * i + j + 1
                if idx >= nrNodes-1:
                    break
                children.append(g.nodes[idx])
            for child in children:
                lvl = parent.privLevel+1
                child.privLevel = lvl
                g.nodesAtLevel[lvl].append(child)
                link(parent, child)

        # add the sink
        g.nodesAtLevel[tree_height].append(g.nodes[-1])
        g.nodes[-1].privLevel = tree_height
        g.nodes[-1].isSink = True
        g.sinks.append(g.nodes[-1])

        # make all leaves point to the sink
        leaves = [n for n in g.nodes if not (n.succ or n.isSink)]
        for l in leaves:
            link(l, g.sinks[0])

        # set a spine. deepest leaf is on the left.
        i = 0
        while i < nrNodes:
            g.nodes[i].isSpine = True
            i = degree*i + 1
        g.nodes[-1].isSpine = True

        if is_disjoint(g):
            raise Exception('Disjoint graph')

        g = finalizeEdges(g)

        return g

    def graph(self):
        return self.__graph()



class uniform(object):
    def __init__(self, _min = None, _max = None):
        self.min = _min if _min else _max
        self.max = _max if _max else _min
    def __call__(self):
        return tg.uniform_int(self.min, self.max)()
    def getadjusted(self, _min, _max):
        return uniform(_min, _max)

class binomial(object):
    def __init__(self, _min = None, _max = None, _p = 0.5):
        self.min = _min if _min else _max
        self.max = _max if _max else _min
        self.p = _p
    def __call__(self):
        n = self.max - self.min
        d = 0
        if n != 0:
            d = dist.binomial(n, self.p)
            while d > self.max - self.min:
                d = dist.binomial(n, self.p)
        return d + self.min
    def getadjusted(self, _min, _max):
        return binomial(_min, _max, self.p)

class geometric(object):
    def __init__(self, _min = None, _max = None, _avg = None):
        self.min = _min if _min else _max
        self.max = _max if _max else _min
        self._avg = _avg
        avg = (_max - _min)/2.0 if not _avg and self.min and self.max else _avg
        self.p = 1.0/(avg + 1.0)
    def __call__(self):
        n = self.max - self.min
        d = 0
        if n != 0:
            d = dist.geometric(self.p)
            while d > self.max - self.min:
                d = dist.geometric(self.p)
        return d + self.min
    def getadjusted(self, _min, _max):
        if self._avg:
            return geometric(_min, _max, self._avg)
        else:
            return geometric(_min, _max)

def main():
    nrNodes = tg.uniform_int(20, 20)
    aspectRatio = tg.uniform(0.5,0.75)
    depth = tg.uniform_int(10, 10)
    width = tg.uniform_int(10, 10)

    level = binomial()
    fan = geometric(1, 3, 1.2)
    oppFanCap = 3
    edge = geometric(1, 3)
    nrSrc = tg.uniform_int(1, 3)
    nrSink = tg.uniform_int(1, 3)
    bottomUp = False
    flip = False

    # generates complex graphs
#    gen = GraphGenerator(nrNodes, depth, level, fan, oppFanCap, edge, nrSrc, nrSink, bottomUp, flip)

    # note no use of fan-out. only the single source has an out-degree >= 1
#    gen = ParallelPipeGraphGenerator(nrNodes, depth)

    # similar to ParallelPipeGraphGenerator, but aspect ratio given instead of depth
#    gen = RectGraphGenerator(nrNodes, aspectRatio)
    gen = RectGraphGenerator2(nrNodes, width)

    # 2 for binary tree, 3 for trinary tree, etc.
#    fan_out = tg.uniform_int(2,2)
#    gen = TreeGraphGenerator(nrNodes, fan_out)

    g = gen.graph()
    f = open('blah.dot', 'w')
    f.write(g.dot())
    f.close()

if __name__ == '__main__':
    main()
