from __future__ import division

from math import trunc, exp, log, ceil, floor

import random

from schedcat.model.LLC_params import *
import schedcat.model.ohc as ohc
import schedcat.model.mcx_tasks as ts

def uniform_int(minval, maxval):
    "Create a function that draws ints uniformly from {minval, ..., maxval}"
    def _draw():
        return random.randint(minval, maxval)
    return _draw

def uniform(minval, maxval):
    "Create a function that draws floats uniformly from [minval, maxval]"
    def _draw():
        return random.uniform(minval, maxval)
    return _draw

def log_uniform(minval, maxval):
    "Create a function that draws floats log-uniformly from [minval, maxval]"
    def _draw():
        return exp(random.uniform(log(minval), log(maxval))) 
    return _draw

def log_uniform_int(minval, maxval):
    "Create a function that draws ints log-uniformly from {minval, ..., maxval}"
    draw_float = log_uniform(minval, maxval + 1)
    def _draw():
        val = int(draw_float())
        val = max(minval, val)
        val = min(maxval, val)
        return val
    return _draw

def uniform_choice(choices):
    "Create a function that draws uniformly elements from choices"
    selector = uniform_int(0, len(choices) - 1)
    def _draw():
        return choices[selector()]
    return _draw

def geometric(p, include_zero = False):
    if include_zero:
        rounder = floor
    else:
        rounder = ceil
    def _draw():
        return rounder(log(1-random.random()) / log(1-p))
    return _draw

def truncate(minval, maxval):
    def _limit(fun):
        def _f(*args, **kargs):
            val = fun(*args, **kargs)
            return min(maxval, max(minval, val))
        return _f
    return _limit

def redraw(minval, maxval):
    def _redraw(dist):
        def _f(*args, **kargs):
            in_range = False
            while not in_range:
                val = dist(*args, **kargs)
                in_range = minval <= val <= maxval
            return val
        return _f
    return _redraw

def exponential(minval, maxval, mean, limiter=redraw):
    """Create a function that draws floats from an exponential
    distribution with expected value 'mean'. If a drawn value is less
    than minval or greater than maxval, then either another value is
    drawn (if limiter=redraw) or the drawn value is set to minval or
    maxval (if limiter=truncate)."""
    def _draw():
        return random.expovariate(1.0 / mean)
    return limiter(minval, maxval)(_draw)

def multimodal(weighted_distributions):
    """Create a function that draws values from several distributions
    with probability according to the given weights in a list of
    (distribution, weight) pairs."""
    total_weight = sum([w for (d, w) in weighted_distributions])
    selector = uniform(0, total_weight)
    def _draw():
        x = selector()
        wsum = 0
        for (d, w) in weighted_distributions:
            wsum += w
            if wsum >= x:
                return d()
        assert False # should never drop off
    return _draw

def restrict(bounds, wMin, wMax):
    interval = bounds[1]-bounds[0]
    return [bounds[0]+wMin*interval, bounds[0]+wMax*interval]

class Custom:
    pass
    
class TaskGeneratorMCX(object):
    """Sporadic task generator"""

    def __init__(self, period, util, util_dist, mem_util, avg_percent, pfuncs, maxWays = 16, maxSets = 16, deadline=lambda x, y: y, ):
        """Creates TaskGenerator based on a given a period and
        utilization distributions."""
        self.period    = period
        self.util      = util
        self.deadline  = deadline
        self.mW = maxWays
        self.mS = maxSets
        self.util_dist = util_dist
        self.mem_util = mem_util
        self.P = pfuncs
        self.avg_percent = avg_percent
    
#     def Pv(self, S, W, Psam):
#         return (1-Psam.kmin)*exp(-S*Psam.ks-W*Psam.kw-S*W*Psam.kw)+Psam.kmin

    def Pv(self, sz, Psam):
        return (1-Psam.kmin)*exp(-sz*Psam.ks)+Psam.kmin
    
    def Psample(self, P):
        Psamp = Custom()
        
        Psamp.ks = P.ks()
        Psamp.kw = P.kw()
        Psamp.kp = P.kp()
        Psamp.kmin = P.kmin()
        return Psamp

    def tasks(self, tot_mc2_util, max_tasks=None, max_util=None, squeeze=False,
              time_conversion=trunc, a_inflate = 1.5, level = 'C'):
        """Generate a sequence of tasks until either max_tasks is reached
        or max_util is reached. If max_util would be exceeded and squeeze is
        true, then the last-generated task's utilization is scaled to exactly
        match max_util. Otherwise, the last-generated task is discarded.
        time_conversion is used to convert the generated (non-integral) values
        into integral task parameters.
        """
        count = 0
        usum  = 0
        while ((max_tasks is None or count < max_tasks) and
               (max_util is None  or usum  < max_util)):
            period   = self.period[level]()
            util     = self.util[level]()
            cost     = period * util
            deadline = self.deadline(cost, period)
            # scale as required
            period   = max(1, int(time_conversion(period)))
            cost     = max(1, int(time_conversion(cost)))
            deadline = max(1, int(time_conversion(deadline)))
            util = cost / period
            count  += 1
            usum   += util
            if max_util and usum > max_util:
                if squeeze:
                    # make last task fit exactly
                    util -= (usum - max_util)
                    cost = trunc(period * util)
                else:
                    break
            
            if cost > 0:
                
                wss = self.mem_util()*cost/(1.5*ohc.CPMD) #of 1-way X 1-color LLC area "units" 
                    
                # ----Generate A and B Partition Functions---- #
                PA = self.Psample(self.P[level]['A'])
                               
                # changed Pv(s, w, PC) to Pv(s*w, PC) and so throughout this block
                scales = {'A': [[self.Pv(min(s*w, wss), PA) for w in range(0, self.mW+1)] for s in range(0, self.mS+1)]}
                
                # First set up A
                for i in range(0, self.mW+1):
                    scales['A'][0][i] = 1.0
                for i in range(0, self.mS+1):
                    scales['A'][i][0] = 1.0
                        
                minA = scales['A'][self.mS][self.mW]
                for s in range(0, self.mS+1):
                    for w in range(0, self.mW+1):
                        scales['A'][s][w] /= minA
            
                # Then B
                scales['B'] = [[scales['A'][s][w]/a_inflate for w in range(0, self.mW+1)] for s in range(0, self.mS+1)]
                
#                 PB = self.Psample(self.P[level]['B'])
#                 
#                 wcscale = 1/self.avg_scale()
#                 scales['B'] = [[max(scales['C'][s][w], wcscale*self.Pv(min(s*w, wss), PB)) for w in range(0, self.mW+1)] for s in range(0, self.mS+1)]
#                 
#                 for w in range(0, self.mW+1):
#                     scales['B'][0][w] = wcscale
#                 for s in range(0, self.mS+1):
#                     scales['B'][s][0] = wcscale
#                         
#                 # Then A
#                 scales['A'] = [[a_inflate*scales['B'][s][w] for w in range(0, self.mW+1)] for s in range(0, self.mS+1)]
#                 
#                 cost /= scales['A'][0][0]
                
                costs = dict()
                
                for key in scales.keys():
                    costs[key] = [[scale*cost for scale in scales_s] for scales_s in scales[key]]
                    
                
                backload = tot_mc2_util - costs['A'][self.mW][self.mS]/period
                
                c_percent = self.avg_percent(backload, level)
                
                costs['Ciso'] = [[costs['B'][s][w]*c_percent for w in range(0, self.mW+1)] for s in range(0, self.mS+1)]
                
                highCcost = costs['Ciso'][0][0]
                lowCcost = costs['Ciso'][self.mS][self.mW]
                
                lowCshared = lowCcost+random.uniform(0.3, 0.7)*(highCcost-lowCcost)
                
                costs['Cshared'] = [[highCcost-(float(s*w)/float(self.mW*self.mS))*(highCcost-lowCshared) for w in range(0, self.mW+1)] for s in range(0, self.mS+1)]
                
                if not level is 'C':
                    costs['C'] = [[costs['Ciso'][s][w] for w in range(0, self.mW+1)] for s in range(0, self.mS+1)]
                
                yield ts.SporadicTaskMCX(exec_cost = costs, period = period, wss = wss, level = level, deadline = deadline)
            else:
                break
                
    def make_task_set(self, max_util):
        a_inf = 1.5
        utilA_A = self.util_dist['A']()
        utilB_A = self.util_dist['B']()
        utilC_A = self.util_dist['C']()
        normFactor = max_util/(utilA_A+utilB_A+utilC_A)
        utilA_A *= normFactor
        utilB_A *= normFactor
        utilC_A *= normFactor
        
        alltasks = []
        alltasks.extend(self.tasks(squeeze = True, tot_mc2_util = max_util, level = 'A', max_util = utilA_A, a_inflate = a_inf))
        alltasks.extend(self.tasks(squeeze = True, tot_mc2_util = max_util, level = 'B', max_util = utilB_A, a_inflate = a_inf))
        alltasks.extend(self.tasks(squeeze = True, tot_mc2_util = max_util, level = 'C', max_util = utilC_A, a_inflate = a_inf))
        
        ts_mcx = ts.TaskSystemMCX(alltasks)
        ts_mcx.a_inflate_choice = a_inf
        
        return ts_mcx
